import tempfile
import textwrap
from services.Service import Service
import os
import random
import copy
import shutil
from google.cloud import pubsub_v1
from google.api_core.exceptions import AlreadyExists

class StorageHandler(): 

    default_conf = {
                    "platform": "linux/amd64",
                        "healthcheck": 
                        {
                            "test": [
                                "CMD-SHELL", "sh", "-c", "nc -z 127.0.0.1 4443 || exit 1"
                            ],"interval": "5s",
                            "timeout": "2s",
                            "retries": 3,
                            "start_period": "1s"
                        },
                        "environment": {
                            "PORT": "4443",
                            "DATA": "/data"
                        },
                        "image": "fsouza/fake-gcs-server",
                        "ports": "{host_port}:4443",
                        "volumes": ["{storage_folder}:/data"]

                    }
    
                                   
    @classmethod
    def _load_default_config(cls, host_port_index, storage_folder):
        host_port = str(4443 + host_port_index)
        # get filename from file path
        config = copy.deepcopy(cls.default_conf)
        config['ports'] = [config['ports'].replace("{host_port}", host_port)]
        for index, volume in enumerate(config['volumes']):
            config['volumes'][index] = config['volumes'][index].replace("{storage_folder}", storage_folder)
            
        return config
    
    @classmethod
    def _create_empty_file(cls, file_path):
        with open(file_path, "w") as file:
            file.write("")
        return file_path
    
    @classmethod        
    def build(cls, service:Service, index: int):     
        random_number = random.randint(1000, 9999)
        temp_dir = tempfile.gettempdir()
        storage_folder = f"/{temp_dir}/{service.name}_{random_number}"
        os.makedirs(storage_folder, exist_ok=True)
        for bucket in service.buckets:
            os.makedirs(f"{storage_folder}/{bucket.name}", exist_ok=True)
            if bucket.location:
                # copy the content of the bucket.location to the bucket folder
                shutil.copytree(bucket.location, f"{storage_folder}/{bucket.name}", 
                                                            dirs_exist_ok=True)
            else:
                # create an empty file in the bucket folder
                cls._create_empty_file(f"{storage_folder}/{bucket.name}/empty_file")                          
            
        return cls._load_default_config(index, storage_folder)             



import os
import random
import time
from datetime import datetime
from time import sleep

import requests
from google.cloud import firestore

expected_site_names = ["24h.com.cy", "balla.com.cy", "stylista.com.cy"]
expected_mo_plan = {
    "cost": 0.0,
    "discount": 0.0,
    "media_owner_id": "nikodea-eqkmy",
    "mo_plan_id": None,
    "packages": [
        {
            "expiration_date": "2032-12-31",
            "is_free": True,
            "package_id": "46e587b9-f130-4c5a-aeed-c6c6901f8386",
            "package_name": "free_trial",
            "subscriptions": [
                {
                    "active": True,
                    "expiration_date": "2032-12-31",
                    "level": "SITE",
                    "pageviews": 0,
                    "payment_date": datetime.now().strftime("%Y-%m-%d"),
                    "site_id": "24hcomcy-e7rjr",
                    "type": "iMOC",
                },
                {
                    "active": True,
                    "expiration_date": "2032-12-31",
                    "level": "SITE",
                    "pageviews": 0,
                    "payment_date": datetime.now().strftime("%Y-%m-%d"),
                    "site_id": "24hcomcy-e7rjr",
                    "type": "BETA_SITE",
                },
                {
                    "active": True,
                    "expiration_date": "2032-12-31",
                    "level": "SITE",
                    "pageviews": 0,
                    "payment_date": datetime.now().strftime("%Y-%m-%d"),
                    "site_id": "ballacomcy-4wrmj",
                    "type": "iMOC",
                },
                {
                    "active": True,
                    "expiration_date": "2032-12-31",
                    "level": "SITE",
                    "pageviews": 0,
                    "payment_date": datetime.now().strftime("%Y-%m-%d"),
                    "site_id": "ballacomcy-4wrmj",
                    "type": "BETA_SITE",
                },
                {
                    "active": True,
                    "expiration_date": "2032-12-31",
                    "level": "SITE",
                    "pageviews": 0,
                    "payment_date": datetime.now().strftime("%Y-%m-%d"),
                    "site_id": "stylistacomcy-c2g1y",
                    "type": "iMOC",
                },
                {
                    "active": True,
                    "expiration_date": "2032-12-31",
                    "level": "SITE",
                    "pageviews": 0,
                    "payment_date": datetime.now().strftime("%Y-%m-%d"),
                    "site_id": "stylistacomcy-c2g1y",
                    "type": "BETA_SITE",
                },
                {
                    "active": True,
                    "expiration_date": "2042-12-29",
                    "level": "MO",
                    "payment_date": datetime.now().strftime("%Y-%m-%d"),
                    "type": "MOC",
                },
                {
                    "active": True,
                    "expiration_date": "2032-12-31",
                    "level": "MO",
                    "payment_date": datetime.now().strftime("%Y-%m-%d"),
                    "type": "BETA_MO",
                },
                {
                    "active": True,
                    "expiration_date": "2033-01-30",
                    "level": "MO",
                    "payment_date": datetime.now().strftime("%Y-%m-%d"),
                    "type": "cMOC",
                },
                {
                    "active": True,
                    "expiration_date": "2033-03-31",
                    "level": "MO",
                    "payment_date": datetime.now().strftime("%Y-%m-%d"),
                    "type": "sMOC",
                },
                {
                    "active": True,
                    "expiration_date": "2032-12-31",
                    "level": "MO",
                    "payment_date": datetime.now().strftime("%Y-%m-%d"),
                    "type": "dMOC",
                },
            ],
        }
    ],
    "updated_at": None,
}

expected_products = [
    {
        "et_data_received": False,
        "id": "c1428634-989e-4d40-ba0b-27825ea65007",
        "media_owner_id": "nikodea-5pi5n",
        "name": "nikodea-5pi5n_24hcomcy-cgfvz",
        "pa_domain_id": 1765,
        "serving_method": "SITE",
        "site_id": "24hcomcy-cgfvz",
        "site_tag": '<script async src="https://moc-sitetags-dev.projectagora.info/sites/24hcomcy-cgfvz/tags/pamoc_loader"></script>',
        "sitetag_domain": "https://moc-sitetags-dev.projectagora.info",
        "status": "ACTIVATED",
    },
    {
        "et_data_received": False,
        "id": "2a821619-6e1e-42bf-9f0d-51efcd7449ec",
        "media_owner_id": "nikodea-5pi5n",
        "name": "nikodea-5pi5n_ballacomcy-f6757",
        "pa_domain_id": 1766,
        "serving_method": "SITE",
        "site_id": "ballacomcy-f6757",
        "site_tag": '<script async src="https://moc-sitetags-dev.projectagora.info/sites/ballacomcy-f6757/tags/pamoc_loader"></script>',
        "sitetag_domain": "https://moc-sitetags-dev.projectagora.info",
        "status": "ACTIVATED",
    },
    {
        "et_data_received": False,
        "id": "742ea2c7-4dbb-4be6-852e-874861ae9e5c",
        "media_owner_id": "nikodea-5pi5n",
        "name": "nikodea-5pi5n_stylistacomcy-9x5mx",
        "pa_domain_id": 1767,
        "serving_method": "SITE",
        "site_id": "stylistacomcy-9x5mx",
        "site_tag": '<script async src="https://moc-sitetags-dev.projectagora.info/sites/stylistacomcy-9x5mx/tags/pamoc_loader"></script>',
        "sitetag_domain": "https://moc-sitetags-dev.projectagora.info",
        "status": "ACTIVATED",
    },
]

expected_firestore_docs = {
    "_24hcomcy-7t9y0": {
        "_0.1": {
            "built_time": 1720537457123,
            "event_tracker_url": "https%3A%2F%2Fevent-tracker-library.mediaownerscloud.com%2Fmain%2F1.0.4%2Fpa_et.min.js",
            "integrations": [{"tag": "", "service": "infra_service"}],
            "core_api_version": "0.0.40",
            "canary_probability": 0,
            "status": "STABLE",
            "event_tracker_version": "1.0.4",
        },
        'serving_method': 'SITE',
        "product_active": True,
    },
    "_ballacomcy-5qanh": {
        "_0.1": {
            "built_time": 1720537457404,
            "event_tracker_url": "https%3A%2F%2Fevent-tracker-library.mediaownerscloud.com%2Fmain%2F1.0.4%2Fpa_et.min.js",
            "integrations": [{"tag": "", "service": "infra_service"}],
            "core_api_version": "0.0.40",
            "canary_probability": 0,
            "status": "STABLE",
            "event_tracker_version": "1.0.4",
        },
        'serving_method': 'SITE',
        "product_active": True,
    },
    "_stylistacomcy-2v2x3": {
        "_0.1": {
            "built_time": 1720537457641,
            "event_tracker_url": "https%3A%2F%2Fevent-tracker-library.mediaownerscloud.com%2Fmain%2F1.0.4%2Fpa_et.min.js",
            "integrations": [{"tag": "", "service": "infra_service"}],
            "core_api_version": "0.0.40",
            "canary_probability": 0,
            "status": "STABLE",
            "event_tracker_version": "1.0.4",
        },
        'serving_method': 'SITE',
        "product_active": True,
    },
}

expected_bucket_object = {
    "kind": "storage#objects",
    "items": [
        {
            "kind": "storage#object",
            "name": "sitetags/MOC_24hcomcy-e9b3x/0.1/site-tag.min.js",
            "id": "prj-p-moc-a2c2-sitetags-emulator/sitetags/MOC_24hcomcy-e9b3x/0.1/site-tag.min.js",
            "bucket": "prj-p-moc-a2c2-sitetags-emulator",
            "size": "1283",
            "contentType": "application/javascript",
            "crc32c": "aeD1ew==",
            "acl": [
                {
                    "bucket": "prj-p-moc-a2c2-sitetags-emulator",
                    "entity": "projectOwner-test-project",
                    "etag": "RVRhZw==",
                    "kind": "storage#objectAccessControl",
                    "object": "sitetags/MOC_24hcomcy-e9b3x/0.1/site-tag.min.js",
                    "projectTeam": {},
                    "role": "OWNER",
                }
            ],
            "md5Hash": "2O950JlV1SxMCxFfQ5YjrA==",
            "etag": "2O950JlV1SxMCxFfQ5YjrA==",
            "storageClass": "STANDARD",
            "timeCreated": "2024-07-09T15:45:26.461597Z",
            "timeStorageClassUpdated": "2024-07-09T15:45:26.461601Z",
            "updated": "2024-07-09T15:45:26.461601Z",
            "generation": "1720539926461601",
            "selfLink": "https://0.0.0.0:4443/storage/v1/b/prj-p-moc-a2c2-sitetags-emulator/o/sitetags%2FMOC_24hcomcy-e9b3x%2F0.1%2Fsite-tag.min.js",
            "mediaLink": "https://0.0.0.0:4443/download/storage/v1/b/prj-p-moc-a2c2-sitetags-emulator/o/sitetags%2FMOC_24hcomcy-e9b3x%2F0.1%2Fsite-tag.min.js?alt=media",
            "metageneration": "1",
        }
    ],
}
# Point to the Firestore emulator
os.environ["FIRESTORE_EMULATOR_HOST"] = "firestore-emulator:8080"
# Specify your project ID here
project_id = "your-project-id"
# Initialize Firestore client
db = firestore.Client(project=project_id)


def read_bucket_objects(site_ids):
    base_url = "http://gcs_storage_proxy:5050"
    bucket_name = "prj-p-moc-a2c2-sitetags-emulator"
    # Construct the URL for listing objects with the folder_name as a prefix
    bucket_objects = {}
    for site_id, site_name in site_ids:
        url = f"{base_url}/storage/v1/b/{bucket_name}/o?prefix=sitetags/MOC_{site_id}"

        # Make the GET request
        response = requests.get(
            url, verify=False
        )  # verify=False may be necessary if using self-signed SSL certificates

        # Parse the response to check if any objects exist with the specified prefix
        if response.status_code == 200:
            data = response.json()
            if "items" in data and len(data["items"]) > 0:
                # print(f"Folder {data}")
                bucket_objects.update({site_id: data})
            else:
                print(f"Folder does not exist in bucket '{bucket_name}'.")
                return False
        else:
            print(
                f"Failed to check folder. Status code: {response.status_code}, Message: {response.text}"
            )
            return False
    return bucket_objects


def read_site_tag_configuration():
    # for site_id, site_name in site_ids:
    # # Reference to the collection
    site_tag_config_ref = db.collection("site_tag_configuration")
    #     print(f'site_tag_dict: {site_tag_dict}', flush=True)
    # Fetch all documents in the collection
    docs = site_tag_config_ref.stream()
    docs_dicts = {}
    # Iterate through the documents and print them (or process as needed)
    for doc in docs:
        # print(f"{doc.id} => {doc.to_dict()}", flush=True)
        docs_dicts.update({doc.id: doc.to_dict()})
    return docs_dicts


def get_mo_by_mo_id_from_sfm(media_owner_id: str) -> (int, dict):
    res = requests.get(f'http://web_gateway/v1.0/media-owners/{media_owner_id}')
    if res.status_code == 200:
        return res.status_code, res.json()
    else:
        return res.status_code, None


def get_mo_id(publisher_id):
    response = requests.get(
        f"http://web_gateway/v1.0/media-owners?publisher_id={publisher_id}"
    )
    # try catch json
    try:
        return response.json()[0]["media_owner_id"]
    except ValueError:  # catches JSON decoding errors
        print("Failed to decode JSON from response", flush=True)
        return False


def get_mo_plan_id(mo_id):
    response = requests.get(f"http://web_gateway/v1.0/media-owners/{mo_id}/mo-plans")
    # try catch json
    try:
        return response.json()[0]["mo_plan_id"]
    except ValueError:  # catches JSON decoding errors
        print("Failed to decode JSON from response", flush=True)
        return False


def get_mo_plan(mo_id):
    response = requests.get(f"http://web_gateway/v1.0/media-owners/{mo_id}/mo-plans")
    # try catch json
    try:
        return response.json()[0]
    except ValueError:  # catches JSON decoding errors
        print("Failed to decode JSON from response", flush=True)
        return False


def delete_mo_plan(mo_id, mo_plan_id):
    response = requests.delete(
        f"http://web_gateway/v1.0/media-owners/{mo_id}/mo-plans/{mo_plan_id}"
    )
    return response.status_code


def delete_mo(mo_id):
    response = requests.delete(f"http://web_gateway/v1.0/media-owners/{mo_id}")
    return response.status_code


def get_sites(mo_id):
    response = requests.get(f"http://web_gateway/v1.0/media-owners/{mo_id}/sites")
    # try catch json
    try:
        return [(site["site_id"], site["name"]) for site in response.json()]
    except ValueError:  # catches JSON decoding errors
        print("Failed to decode JSON from response", flush=True)
        return False


def get_products(mo_id):
    response = requests.get(f"http://web_gateway/v1.0/media-owners/{mo_id}/products")
    # try catch json
    try:
        return response.json()
    except ValueError:  # catches JSON decoding errors
        print("Failed to decode JSON from response", flush=True)
        return False


def delete_site(mo_id, site_id):
    response = requests.delete(
        f"http://web_gateway/v1.0/media-owners/{mo_id}/sites/{site_id}"
    )
    return response.status_code


def add_site(mo_id):
    site = {
        "domain": "testaro.gr",
        "name": "testaro",
        "pa_domain_id": 2,
        "site_id": "site_id_testaro",
    }
    headers = {"accept": "application/json", "Content-Type": "application/json"}
    response = requests.post(
        f"http://web_gateway/v1.0/media-owners/{mo_id}/sites",
        headers=headers,
        json=site,
    )
    return response.status_code


def mo_migrate(publisher_id, dry_run=False):
    headers = {"accept": "application/json", "Content-Type": "application/json"}
    response = requests.post(
        "http://web_gateway/v1.0/media-owners/migrate",
        headers=headers,
        json={"publisher_id": str(publisher_id), "dry_run": dry_run},
    )
    print("end of MO MIGRATE", flush=True)
    return response.status_code


def prepare_expected_mo_plan(mo_plan, site_ids):
    expected_mo_plan["updated_at"] = mo_plan["updated_at"]
    expected_mo_plan["mo_plan_id"] = mo_plan["mo_plan_id"]
    expected_mo_plan["media_owner_id"] = mo_plan["media_owner_id"]
    for index, subscription in enumerate(
            expected_mo_plan["packages"][0]["subscriptions"]
    ):
        for site_id, site_name in site_ids:
            if "site_id" in subscription:
                if subscription["site_id"].split("-")[0] == site_id.split("-")[0]:
                    expected_mo_plan["packages"][0]["subscriptions"][index][
                        "site_id"
                    ] = site_id
    return expected_mo_plan


def prepare_expected_products(products, site_ids):
    for index, product in enumerate(products):
        for site_id, site_name in site_ids:
            if expected_products[index]["site_id"].split("-")[0] == site_id.split("-")[0]:
                expected_products[index]["site_id"] = site_id
                expected_products[index][
                    "name"
                ] = f"{product['media_owner_id']}_{site_id}"
                expected_products[index][
                    "site_tag"
                ] = f'<script async src="https://moc-sitetags-dev.projectagora.info/sites/{site_id}/tags/pamoc_loader"></script>'
        for index, expected_product in enumerate(expected_products):
            if expected_product["site_id"].split("-")[0] == product["site_id"].split("-")[0]:
                expected_products[index]["id"] = product["id"]
                expected_products[index]["media_owner_id"] = product["media_owner_id"]
    return expected_products


def assert_nodes():
    mo_id = get_mo_id(6)
    site_ids = get_sites(mo_id)
    print(f"assert_nodes site_ids:{site_ids}", flush=True)
    for site_id, site_name in site_ids:
        assert (
                site_name in expected_site_names
        ), f"Expected site name {site_name} not found"
    mo_plan = get_mo_plan(mo_id)
    exp_mo_plan = prepare_expected_mo_plan(mo_plan, site_ids)
    assert mo_plan == exp_mo_plan, f"Expected mo_plan {exp_mo_plan} not found"
    time.sleep(5)
    print("assert_nodes WAKE UP from sleep", flush=True)
    products = get_products(mo_id)
    print(f"products:{products}", flush=True)
    exp_products = prepare_expected_products(products, site_ids)
    assert sorted(products, key=lambda x: x['id']) == sorted(exp_products, key=lambda x: x['id']), f"Expected products {exp_products} not found"
    firestore_docs = read_site_tag_configuration()
    print(f"firestore_docs:{firestore_docs}", flush=True)
    bucket_objects = read_bucket_objects(site_ids)
    print(f"bucket_objects:{bucket_objects}", flush=True)
    for site_id, site_name in site_ids:
        assert (
                f"_{site_id}" in firestore_docs
        ), f"Expected firestore doc {site_id} not found"
        expected_firestore_docs["_24hcomcy-7t9y0"]["_0.1"]["built_time"] = (
            firestore_docs[f"_{site_id}"]["_0.1"]["built_time"]
        )
        print(f"expected_firestore_docs:{expected_firestore_docs}", flush=True)
        assert (
                firestore_docs[f"_{site_id}"] == expected_firestore_docs["_24hcomcy-7t9y0"]
        ), f"firestore doc {site_id} not equal to expected value"
        assert (
                site_id in bucket_objects
        ), f"Expected bucket object {site_id} not found"
        assert (
                bucket_objects[site_id]["items"][0]["name"]
                == f"sitetags/MOC_{site_id}/0.1/site-tag.min.js"
        ), f"Expected bucket object {site_id} not found"
        assert (
                bucket_objects[site_id]["items"][0]["id"]
                == f"prj-p-moc-a2c2-sitetags-emulator/sitetags/MOC_{site_id}/0.1/site-tag.min.js"
        ), f"Expected bucket object {site_id} not found"
        assert (
                bucket_objects[site_id]["items"][0]["acl"][0]["object"]
                == f"sitetags/MOC_{site_id}/0.1/site-tag.min.js"
        ), f"Expected bucket object {site_id} not found"
        print(f"site_id, site_name:{site_id}, {site_name}", flush=True)

def test_no_mo_exist_migrate():
    print("TEST1: test_no_mo_exist_migrate", flush=True)

    mo_id = get_mo_id(6)
    print(f"TEST1: mo_id={mo_id}", flush=True)
    if mo_id:
        mo_pla_id = get_mo_plan_id(mo_id)
        if mo_pla_id:
            delete_mo_plan(mo_id, mo_pla_id)
        delete_mo(mo_id)
    print("TEST1: start of mo migrate", flush=True)    
    status_code = mo_migrate(6, dry_run=False)
    print("TEST1: end of  mo migrate", flush=True)
    assert status_code == 201, f"Expected status code 201, got {status_code}"
    assert_nodes()


def test_mo_exist_no_sites_migrate():
    print("TEST2: test_mo_exist_no_sites_migrate", flush=True)
    mo_id = get_mo_id(6)
    if mo_id:
        site_ids = get_sites(mo_id)
        if site_ids:
            for site_id, site_name in site_ids:
                status_code = delete_site(mo_id, site_id)
                print(status_code, flush=True)
    status_code = mo_migrate(6, dry_run=False)
    assert status_code == 201, f"Expected status code 201, got {status_code}"
    assert_nodes()


def test_mo_exist_one_more_site_migrate():
    print("TEST3: test_mo_exist_one_more_site_migrate", flush=True)
    mo_id = get_mo_id(6)
    if mo_id:
        add_site(mo_id)
    status_code = mo_migrate(6, dry_run=False)
    assert status_code == 201, f"Expected status code 201, got {status_code}"
    assert_nodes()


def create_mo(mo_id: str, mo_name: str, pa_id: int) -> (int, dict):
    new_mo = {
        "media_owner_id": mo_id,
        "name": mo_name,
        "sites": [],
        "company_details": {
            "name": 'company_name',
            "type": "Company",
            "vat_registration_office": "1423424234",
            "registration_number": "1423424234",
            "country": "Greece",
            "address": "14o xlm E.O. Athinon - Lamias, 14564, Kifisia",
            "city": "Athens",
            "zip_code": "12345",
        },
        "contact_details": {
            "first_name": "Sofia",
            "last_name": "Kontopyraki",
            "email": "s.kontopyraki@dpg.gr",
            "phone": "+306947435188",
        },
        "currency": "EUR",
        "pa_id": pa_id,
    }
    response = requests.post(f"http://web_gateway/v1.0/media-owners", json=new_mo)
    if response.status_code == 201:
        return response.status_code, response.json()
    else:
        return response.status_code, None


def create_a_new_site(mo_id: str, site_domain: str, site_name: str, site_pa_domain_id: int, site_id: str) -> (int, dict):
    new_site = {
        "domain": site_domain,
        "name": site_name,
        "pa_domain_id": site_pa_domain_id,
        "site_id": site_id,
    }
    headers = {"accept": "application/json", "Content-Type": "application/json"}
    response = requests.post(
        f"http://web_gateway/v1.0/media-owners/{mo_id}/sites",
        headers=headers,
        json=new_site,
    )
    if response.status_code == 201:
        return response.status_code, response.json()
    else:
        return response.status_code, None


def create_a_new_mo_plan(mo_id, site_id) -> (int, dict):
    new_mo_plan = {
        "cost": 0.0,
        "discount": 0.0,
        "media_owner_id": mo_id,
        # "mo_plan_id": None,
        "packages": [
            {
                "expiration_date": "2032-12-31",
                "is_free": True,
                "package_id": "46e587b9-f130-4c5a-aeed-c6c6901f8386",
                "package_name": "free_trial",
                "subscriptions": [
                    {
                        "active": True,
                        "expiration_date": "2032-12-31",
                        "level": "SITE",
                        "pageviews": 0,
                        "payment_date": datetime.now().strftime("%Y-%m-%d"),
                        "site_id": site_id,
                        "type": "iMOC",
                    },
                    {
                        "active": True,
                        "expiration_date": "2032-12-31",
                        "level": "SITE",
                        "pageviews": 0,
                        "payment_date": datetime.now().strftime("%Y-%m-%d"),
                        "site_id": site_id,
                        "type": "BETA_SITE",
                    },
                    {
                        "active": True,
                        "expiration_date": "2042-12-29",
                        "level": "MO",
                        "payment_date": datetime.now().strftime("%Y-%m-%d"),
                        "type": "MOC",
                    },
                    {
                        "active": True,
                        "expiration_date": "2032-12-31",
                        "level": "MO",
                        "payment_date": datetime.now().strftime("%Y-%m-%d"),
                        "type": "BETA_MO",
                    },
                    {
                        "active": True,
                        "expiration_date": "2033-01-30",
                        "level": "MO",
                        "payment_date": datetime.now().strftime("%Y-%m-%d"),
                        "type": "cMOC",
                    },
                    {
                        "active": True,
                        "expiration_date": "2033-03-31",
                        "level": "MO",
                        "payment_date": datetime.now().strftime("%Y-%m-%d"),
                        "type": "sMOC",
                    },
                    {
                        "active": True,
                        "expiration_date": "2032-12-31",
                        "level": "MO",
                        "payment_date": datetime.now().strftime("%Y-%m-%d"),
                        "type": "dMOC",
                    },
                ],
            }
        ],
        "updated_at": None,
    }
    res = requests.post(f'http://web_gateway/v1.0/media-owners/{mo_id}/mo-plans', json=new_mo_plan)
    if res.status_code == 201:
        return res.status_code, res.json()
    else:
        return res.status_code, None


def test_create_a_new_media_owner():
    print("TEST4: test_create_a_new_media_owner", flush=True)
    new_mo_id = f'media_owner_id_{str(random.randint(0, 1000))}'
    new_mo_name = f'media_owner_name_{str(random.randint(0, 1000))}'
    new_mo_pa_id = random.randint(0, 1000)
    status_code, mo = get_mo_by_mo_id_from_sfm(media_owner_id=new_mo_id)
    assert status_code != 200
    assert mo is None
    print(f'Media owner with id:{new_mo_id} actually does not exist initially', flush=True)
    status_code, mo = create_mo(mo_id=new_mo_id, mo_name=new_mo_name, pa_id=new_mo_pa_id)
    assert status_code == 201
    assert mo is not None
    print(f'Media owner with id:{new_mo_id} was created', flush=True)
    status_code, mo = get_mo_by_mo_id_from_sfm(media_owner_id=new_mo_id)
    assert status_code == 200
    assert mo is not None
    print(f'Media owner with id:{new_mo_id} now actually exists in the database', flush=True)

    # Add a site
    site_domain = f'site_domain_{str(random.randint(0, 1000))}'
    site_name = f'site_name_{str(random.randint(0, 1000))}'
    site_pa_domain_id = random.randint(0, 1000)
    site_id = f'site_id_{str(random.randint(0, 1000))}'
    status_code, _site = create_a_new_site(mo_id=new_mo_id, site_domain=site_domain, site_name=site_name, site_pa_domain_id=site_pa_domain_id, site_id=site_id)
    assert status_code == 201
    assert _site is not None
    print(f'Added a site with id:{site_id} to MO:{new_mo_id}', flush=True)
    status_code, mo = get_mo_by_mo_id_from_sfm(media_owner_id=new_mo_id)
    assert status_code == 200
    assert len(mo['sites']) == 1
    assert mo['sites'][0]['site_id'] == site_id
    print(f'Site with id:{site_id} actually exists in the DB', flush=True)

    # Create a MO-Plan for MO
    get_products_response = get_products(mo_id=new_mo_id)
    assert get_products_response is False
    print(f'Media Owner:{new_mo_id} does not currently have a Product', flush=True)
    status_code, mo_plan = create_a_new_mo_plan(mo_id=new_mo_id, site_id=site_id)
    assert status_code == 201
    assert mo_plan is not None
    print(f'Created a Mo-Plan for MO:{new_mo_id}', flush=True)
    time.sleep(30)

    # check if SM created a product
    get_products_response = get_products(mo_id=new_mo_id)
    assert get_products_response is not False
    print(f'{get_products_response=}', flush=True)
    assert isinstance(get_products_response, list)
    assert len(get_products_response) == 1
    print(f'Media Owner:{new_mo_id} now has a Product', flush=True)


if __name__ == "__main__":
    print("TESTS STARTED", flush=True)
    test_no_mo_exist_migrate()
    test_mo_exist_no_sites_migrate()
    test_mo_exist_one_more_site_migrate()
    test_create_a_new_media_owner()
    print("FINISHED", flush=True)
    sleep(5)



import click
from transports.cli.v1_0.commands.project_commands import create_project, set_active_project, list_projects, run_test, delete_project, update_project

@click.group()
def cli():
    pass

cli.add_command(create_project)
cli.add_command(update_project)
cli.add_command(set_active_project)
cli.add_command(list_projects)
cli.add_command(delete_project)
cli.add_command(run_test)


if __name__ == "__main__":
    cli()



import argparse
import json
from subprocess import Popen, PIPE, STDOUT

from platforms.docker_compose.DockerComposePlatform import DCPlatform
from services.Service import Service
from test_suites.MicroTestSuite import MicroTestSuite


def load_service_config(service_name):
    with open(f'data/services/{service_name}.json') as f:
        service_config = json.load(f)
    # Save the updated configuration back to the file if necessary
    # with open(f'data/services/{service_name}.json', 'w') as f:
    #     json.dump(service_config, f, indent=4)
    return service_config


def update_specific_field(d, keys, value):
    """
    Update a specific field in a nested dictionary without replacing entire sub-dictionaries.
    
    :param d: The dictionary to update.
    :param keys: A list of keys that leads to the specific field to update.
    :param value: The new value to set for the specified field.
    """
    current = d
    for key in keys[:-1]:  # Navigate to the sub-dictionary containing the target field
        current = current.setdefault(key, {})
    current[keys[-1]] = value  # Update the target field


def extract_key_value_pairs(d, current_path=None):
    """
    Recursively extract all key paths and their corresponding values from a nested dictionary.

    :param d: The dictionary to traverse.
    :param current_path: The current path of keys leading to a specific value.
    :return: Yields tuples of (path, value) where path is a list of keys leading to the value.
    """
    if current_path is None:
        current_path = []
    for key, value in d.items():
        # Update the path with the current key
        new_path = current_path + [key]
        if isinstance(value, dict):
            # If the value is a dictionary, continue the recursion
            yield from extract_key_value_pairs(value, new_path)
        else:
            # If the value is not a dictionary, yield the path and value
            yield (new_path, value)


def configure_services_from_config(config_path):
    with open(config_path) as f:
        config = json.load(f)
    services = []
    for service_name, config_updates in config.items():
        service_config = load_service_config(service_name)
        for path, value in extract_key_value_pairs(config_updates):
            update_specific_field(service_config, path, value)
            print(f"Path: {path}, Value: {value}")
        services.append(Service(**service_config))
    return services


def get_identity_token():
    try:
        p = Popen("gcloud auth print-identity-token", shell=True, stdin=PIPE, stdout=PIPE, stderr=STDOUT, close_fds=True)
        auth_token = p.stdout.read().decode("utf-8").strip("\n")
        return auth_token
    except Exception as e:
        print(f"Error while getting identity token: {str(e)}")
        return None


def update_config_with_identity_token(file_path, identity_token):
    # Step 1: Read the JSON file and create a Python dict
    with open(file_path, 'r') as file:
        config = json.load(file)

    # Step 2: Assuming get_identity_token is a function that returns the new identity token
    # it = get_identity_token() # Uncomment and define get_identity_token() as needed

    # Step 3: Update the "web_gateway" -> "environment" -> "X_PA_GCP_JWT" with the new identity token
    config["web_gateway"]["environment"]["X_PA_GCP_JWT"] = identity_token

    # Write the updated configuration back to the file
    with open(file_path, 'w') as file:
        json.dump(config, file, indent=4)


# Example usage
# update_config_with_identity_token('path/to/config.json', 'new_identity_token_here')

def main():
    # Set up argument parsing
    parser = argparse.ArgumentParser(description='Run tests based on a JSON configuration.')
    parser.add_argument('-f', '--file', required=False, default='data/config.json', help='Path to the JSON file with service configurations.')
    parser.add_argument('-t', '--test-path', required=False, default='data/tests', help='Path to the directory containing test definitions.')

    args = parser.parse_args()
    identity_token = get_identity_token()
    update_config_with_identity_token(args.file, identity_token)
    # Load the services by reading the data from the JSON file specified in command line
    services = configure_services_from_config(args.file)

    dc_platform = DCPlatform(yaml_input=None)

    test_suite = MicroTestSuite(id="1", name="Test Suite 1", services=services, platform=dc_platform, tests_path=args.test_path)
    test_suite.run()
    results = test_suite.test_results()
    test_suite.stop()
    if results:
        print("Test results: SUCCESS")
    else:
        print("Test results: FAILURE")


if __name__ == "__main__":
    main()



import tempfile
import textwrap
from services.Service import Service
import os
import random
import copy
import shutil
from google.cloud import pubsub_v1
from google.api_core.exceptions import AlreadyExists

class PubSubHandler(): 

    default_conf = {
                        "platform": "linux/amd64",
                        "image": "google/cloud-sdk",
                        "healthcheck": 
                        {
                            "test": [
                                "CMD-SHELL", "sh", "-c", "nc -z 127.0.0.1 8085 || exit 1"
                            ],"interval": "5s",
                            "timeout": "2s",
                            "retries": 3,
                            "start_period": "1s"
                        },
                        "environment": {
                                "PUBSUB_PROJECT_ID": "your-project-id"
                            },
                        "ports": "{host_port}:8085",
                        "command": "gcloud beta emulators pubsub start --host-port=0.0.0.0:8085"
                        }
    
    pub_sub_init_config = { 
                            "platform": "linux/amd64",
                            "depends_on": "{pub_sub_emulator_host}",
                            "build": {
                                            "context": "{code_folder}",
                                            "dockerfile": "dockerfile.docker"
                                            },
                            "environment": {
                                "PUBSUB_EMULATOR_HOST": "pub_sub_emulator"
                            }
                    
                            }
    
    docker_file_content = """
# Use the google/cloud-sdk base image
FROM google/cloud-sdk

# Install Python and pip
RUN apt-get update && apt-get install -y python3 python3-pip

# Install google-cloud-pubsub
RUN pip3 install google-cloud-pubsub

# Copy local code to the container image.
COPY ./* .

# Set the entrypoint to run the initialization script
ENTRYPOINT ["python3", "/init.py"]
"""

    init_script_content = """from google.cloud import pubsub_v1
import os
import time

# Set the Pub/Sub emulator host
os.environ["PUBSUB_EMULATOR_HOST"] = "pub_sub_emulator:8085"
PROJECT_ID = "your-project-id"

publisher = pubsub_v1.PublisherClient()
subscriber = pubsub_v1.SubscriberClient()

topics_subscriptions = {
    # "topic1": [("subscription1", "http://example.com/push1"), ("subscription2", "http://example.com/push2")],
    # "topic2": [("subscription3", "http://example.com/push3"), ("subscription4", "http://example.com/push4")]
    {TOPICS_SUBSCRIPTIONS}
}

for topic, subscriptions in topics_subscriptions.items():
    topic_path = publisher.topic_path(PROJECT_ID, topic)
    publisher.create_topic(request={"name": topic_path})
    print(f"Topic created: {topic}")

    for subscription_name, push_endpoint in subscriptions:
        subscription_path = subscriber.subscription_path(PROJECT_ID, subscription_name)
        push_config = pubsub_v1.types.PushConfig(push_endpoint=push_endpoint)
        subscriber.create_subscription(
            request={
                "name": subscription_path,
                "topic": topic_path,
                "push_config": push_config
            }
        )
        print(f"Subscription created: {subscription_name} with push endpoint: {push_endpoint}")
#time.sleep(60)        
"""
                           
    
                                   
    @classmethod
    def _load_default_config(cls, host_port_index):
        host_port = str(8185 + host_port_index)
        # get filename from file path
        config = copy.deepcopy(cls.default_conf)
        config['ports'] = [config['ports'].replace("{host_port}", host_port)]
        config['command'] = config['command'].replace("{host_port}", host_port)
            
        return config
    
    @classmethod
    def _load_initiator_config(cls, service_name, service_folder):
        
        # get filename from file path
        config = copy.deepcopy(cls.pub_sub_init_config)
        config['depends_on'] = [config['depends_on'].replace("{pub_sub_emulator_host}", service_name)]
        config['build']['context'] = service_folder
        
            
        return config
    
    @classmethod
    def _create_init_file(cls, service_folder, init_file_content):
       os.makedirs(service_folder, exist_ok=True)
       #init_file_content = init_file_content.replace("{code_folder}", service_folder) 
       #docker_file_content = textwrap.dedent(docker_file_content)  # Remove leading spaces
       with open(f"{service_folder}/init.py", "w") as file:
            file.write(init_file_content)    
            file_path = file.name
       os.chmod(file_path, 0o755)  # Set executable permissions
       return file_path    
   
    @classmethod
    def _create_docker_file(cls, code_folder, docker_file_content):
       os.makedirs(code_folder, exist_ok=True)
       docker_file_content = docker_file_content.replace("{code_folder}", code_folder) 
       docker_file_content = textwrap.dedent(docker_file_content)  # Remove leading spaces
       with open(f"{code_folder}/dockerfile.docker", "w") as file:
            file.write(docker_file_content)    
            file_path = file.name
       os.chmod(file_path, 0o755)  # Set executable permissions
       return file_path 

    
    # @classmethod
    # def create_topic_and_subscriptions(cls, project_id, topic):
        
    #     os.environ["PUBSUB_EMULATOR_HOST"] = "pub_sub_emulator:8085"
    #     print(f"after env set {topic.name}")
    #     # Create a Pub/Sub client
    #     publisher = pubsub_v1.PublisherClient()
    #     subscriber = pubsub_v1.SubscriberClient()

    #     # Create a topic
    #     topic_path = publisher.topic_path(project_id, topic.name)
    #     try:
    #         topic = publisher.create_topic(request={"name": topic_path})
    #         print(f"Topic created: {topic.name}")
    #     except AlreadyExists:
    #         print(f"Topic already exists: {topic_path}")
        
    #     for subscription in topic.subscriptions:
    #         push_config = pubsub_v1.types.PushConfig(push_endpoint=subscription.push_config.push_endpoint)
    #         # Create a subscription to the topic
    #         subscription_path = subscriber.subscription_path(project_id, subscription.name)
    #         subscription = subscriber.create_subscription(
    #             request={"name": subscription_path, "topic": topic.name, "push_config": push_config}
    #         )
    #         print(f"Subscription created: {subscription.name}", flush=True)
    @classmethod
    def create_topics_and_subscriptions(cls, service):
        topics_subscriptions_str = ""
        topics = service.topics
        for topic in topics:
            topics_subscriptions_str += f"    '{topic.name}': ["
            for subscription in topic.subscriptions:
                topics_subscriptions_str += f"('{subscription.name}', '{subscription.push_endpoint}'), "
            # remove trailing comma
            #topics_subscriptions_str = topics_subscriptions_str[:-2]    
            topics_subscriptions_str += "],\n"
        print(f"topics_subscriptions_str: {topics_subscriptions_str}")
        return topics_subscriptions_str
        
       
    @classmethod        
    def build(cls, project_name, service:Service, index: int):                              
            
        return cls._load_default_config(index)             

    @classmethod        
    def build_initiator(cls, project_name, service:Service, index: int):     
        topics_subscriptions = cls.create_topics_and_subscriptions(service)
        init_script_content = cls.init_script_content.replace("{TOPICS_SUBSCRIPTIONS}", topics_subscriptions)
        random_number = random.randint(1000, 9999)
        temp_dir = tempfile.gettempdir()
        service_folder = f"/{temp_dir}/{service.name}_{random_number}"                         
        cls._create_init_file(service_folder, init_script_content)            
        cls._create_docker_file(service_folder, cls.docker_file_content)
        return cls._load_initiator_config(service.name, service_folder)             




import tempfile
from services.Service import Service
import os
import random
import textwrap
import copy
import git

class FunctionHandler(): 

    default_conf = {
                        "platform": "linux/amd64",
                            "healthcheck": 
                            {
                                "test": [
                                    "CMD-SHELL", "sh", "-c", "nc -z 127.0.0.1 8080 || exit 1"
                                ],"interval": "5s",
                                "timeout": "2s",
                                "retries": 3,
                                "start_period": "1s"
                            },
                            "build": {
                                            "context": "{code_folder}",
                                            "dockerfile": "dockerfile.docker"
                                            },
                            "ports": "{host_port}:8080"
                        }
    
    
    
    @classmethod
    def _create_docker_file(cls, code_folder, docker_file_content):
       os.makedirs(code_folder, exist_ok=True)
       docker_file_content = docker_file_content.replace("{code_folder}", code_folder) 
       docker_file_content = textwrap.dedent(docker_file_content)  # Remove leading spaces
       with open(f"{code_folder}/dockerfile.docker", "w") as file:
            file.write(docker_file_content)    
            file_path = file.name
       os.chmod(file_path, 0o755)  # Set executable permissions
       return file_path    

            
                                
    @classmethod
    def _load_default_config(cls, host_port_index, code_folder, temp_sh_file):
        host_port = str(6000 + host_port_index)
        # get filename from file path
        config = copy.deepcopy(cls.default_conf)
        config['ports'] = [config['ports'].replace("{host_port}", host_port)]
        config['build']['context'] = code_folder
            
        return config
    
    
    @classmethod        
    def _common_build(cls, service: Service, index: int, code_folder: str, docker_file_path: str):     
        service_dict = {}
         
        service_dict = cls._load_default_config(index, code_folder, docker_file_path)    
        service_dict['environment'] = service.env_vars
        if service.depends_on_services:
            depends_on = {}
            env_update = {}
            for dependency in service.depends_on_services:
                depends_on.update({
                                    dependency.name : {
                                        "condition": "service_healthy"
                                    }
                                    })
                if dependency.type == 'GCP' and dependency.service_type == 'FIRESTORE':
                    env_update.update({
                        "FIRESTORE_EMULATOR_HOST": f"{dependency.name}:8080"
                    })
                elif dependency.type == 'GCP' and dependency.service_type == 'PUBSUB':
                    env_update.update({
                        "PUBSUB_EMULATOR_HOST": f"{dependency.name}:8085"
                    })
                elif dependency.type == 'GCP' and dependency.service_type == 'STORAGE':
                    env_update.update({
                        "STORAGE_EMULATOR_HOST": "http://gcs_storage_proxy:5050"
                    })        
            service_dict["depends_on"] = depends_on
            service_dict["environment"].update(env_update)
        return service_dict         


class NodeFunctionHandler(FunctionHandler):                
    
    docker_file_content = """
# Use the official lightweight Node.js 14 image.
# https://hub.docker.com/_/node
FROM node:14-slim

# Create and change to the app directory.
WORKDIR /usr/src/app

# Copy application dependency manifests to the container image.
# A wildcard is used to ensure both package.json AND package-lock.json are copied.
# Copying this separately prevents re-running npm install on every code change.
COPY ./package*.json ./

# Install production dependencies.
RUN npm install --only=production

# Install Functions Framework
RUN npm install @google-cloud/functions-framework

# Copy local code to the container image.
COPY . .

# Use the Functions Framework to serve the function.
CMD ["npx", "functions-framework", "--target=main"]
"""

   
    @classmethod        
    def build(cls, service: Service, index: int):     
        service_dict = {}
        
        random_number = random.randint(1000, 9999)
        temp_dir = tempfile.gettempdir()
        code_repo_folder = f"/tmp/{service.name}_{random_number}"
        if service.code_folder:
            code_folder = f"{code_repo_folder}/{service.code_folder}"
        else:
            code_folder = code_repo_folder    
        # git clone the code repo
        git.Repo.clone_from(service.code_repo, code_repo_folder, branch=service.code_branch)
        docker_file_path = cls._create_docker_file(code_folder, cls.docker_file_content)    
        service_dict = cls._common_build(service, index, code_folder, docker_file_path)
        
        return service_dict  

    
    


from abc import ABC, abstractmethod
from typing import List
from services.Service import Service

class IPlatform(ABC): #pragma: no cover

    @abstractmethod
    def build(self, services: List[Service]):
        pass

    @abstractmethod
    def run(self):
        pass



from services.Service import Service
import os
import tempfile
import textwrap
import copy

class PostgresHandler(): 

    default_conf = {
                    "platform": "linux/amd64",
                        "healthcheck": 
                        {
                            "test": [
                                "CMD-SHELL", "sh", "-c", "nc -z 127.0.0.1 5432 || exit 1"
                            ],"interval": "5s",
                            "timeout": "2s",
                            "retries": 3,
                            "start_period": "1s"
                        },
                        "environment": {
                            "POSTGRES_USER": "postgres",
                            "POSTGRES_PASSWORD": "123456!",
                            "POSTGRES_DB": "{db_name}"
                        },
                        "image": "postgres:13.5",
                        "ports": "{host_port}:5432",
                        "volumes": ["/var/lib/postgresql/13/main/{db_name}:/var/lib/postgresql",
                                    "{sql_init_file_path}:/docker-entrypoint-initdb.d/{sql_init_file}",
                                    "{temp_sh_file}:/docker-entrypoint-initdb.d/init-schema.sh"]

                    }
    sh_file_content = """
#!/bin/bash
set -e

# The database name and user are obtained from the environment variables.
DB_NAME=$POSTGRES_DB
DB_USER=$POSTGRES_USER

# Execute the SQL schema file.
echo "Initializing the database schema..."
psql -v ON_ERROR_STOP=1 --username "$DB_USER" --dbname "$DB_NAME" <<-EOSQL
    \i '/docker-entrypoint-initdb.d/{sql_init_file}'
EOSQL

echo "Database schema initialized successfully."
"""
    @classmethod
    def create_init_db_sh_file(cls, init_sql_file):
       sh_file_content = cls.sh_file_content.replace("{sql_init_file}", init_sql_file) 
       sh_file_content = textwrap.dedent(sh_file_content)  # Remove leading spaces
       with tempfile.NamedTemporaryFile(delete=False, suffix=".sh") as temp_file:
            temp_file.write(sh_file_content.encode())
            temp_file_path = temp_file.name
       os.chmod(temp_file_path, 0o755)  # Set executable permissions
       return temp_file_path
            
                                
    @classmethod
    def load_default_config(cls, service_name, host_port_index, init_sql_file_path, temp_sh_file):
        host_port = str(5432 + host_port_index)
        # get filename from file path
        init_sql_file = os.path.basename(init_sql_file_path)
        config = copy.deepcopy(cls.default_conf)
        config['ports'] = [config['ports'].replace("{host_port}", host_port)]
        config['environment']['POSTGRES_DB'] = config['environment']['POSTGRES_DB'].replace("{db_name}", service_name)
        for index, volume in enumerate(config['volumes']):
            config['volumes'][index] = config['volumes'][index].replace("{db_name}", service_name)
            config['volumes'][index] = config['volumes'][index].replace("{sql_init_file_path}", init_sql_file_path)
            config['volumes'][index] = config['volumes'][index].replace("{sql_init_file}", init_sql_file)
            config['volumes'][index] = config['volumes'][index].replace("{temp_sh_file}", temp_sh_file)
            
        return config
    
    
    @classmethod        
    def build(cls, service: Service, index: int):     
        service_dict = {}
        init_sql_file_path = service.sql_init_file
        init_sql_file = os.path.basename(init_sql_file_path)
        temp_sh_file = cls.create_init_db_sh_file(init_sql_file)
        service_dict = cls.load_default_config(service.name, index, init_sql_file_path, temp_sh_file)
        
        return service_dict                    
    


import tempfile
import textwrap
from services.Service import Service
import os
import random
import copy
import shutil
from google.cloud import pubsub_v1
from google.api_core.exceptions import AlreadyExists

class FirestoreHandler(): 

    default_conf = {
                        "platform": "linux/amd64",
                        "image": "google/cloud-sdk",
                        "healthcheck": 
                        {
                            "test": [
                                "CMD-SHELL", "sh", "-c", "nc -z 127.0.0.1 8080 || exit 1"
                            ],"interval": "5s",
                            "timeout": "2s",
                            "retries": 3,
                            "start_period": "1s"
                        },
                        "environment": {
                            "CLOUDSDK_CORE_PROJECT": "your-project-id"
                        },
                        "ports": "{host_port}:8080",
                        "command": "gcloud beta emulators firestore start --host-port=0.0.0.0:8080"
                        }
    
                                   
    @classmethod
    def _load_default_config(cls, host_port_index):
        host_port = str(8090 + host_port_index)
        # get filename from file path
        config = copy.deepcopy(cls.default_conf)
        config['ports'] = [config['ports'].replace("{host_port}", host_port)]
            
        return config
    
    
    @classmethod        
    def build(cls, index: int):     
        return cls._load_default_config(index)             


from time import time
import yaml
import json
from typing import List, Optional
import subprocess
from pydantic import BaseModel
from platforms.IPlatform import IPlatform
from platforms.docker_compose.service_handlers.FunctionHandler import NodeFunctionHandler
from platforms.docker_compose.service_handlers.gcp_services.FirestoreHandler import FirestoreHandler
from platforms.docker_compose.service_handlers.gcp_services.PubSubHandler import PubSubHandler
from platforms.docker_compose.service_handlers.gcp_services.StorageHandler import StorageHandler
from platforms.docker_compose.service_handlers.gcp_services.StorageProxy import StorageProxy
from platforms.docker_compose.service_handlers.PostgresHandler import PostgresHandler
from platforms.docker_compose.service_handlers.TestHandler import PythonTestHandler
from services.Service import Service
import os
from platforms.docker_compose.service_handlers.DockerHandler import DockerHandler

class DockerComposePlatform(BaseModel, IPlatform): #pragma: no cover

    yaml_input: Optional[str] = None
    test_results: Optional[bool] = None
    
    def validate_yaml_input(self) -> bool:
        try:
            yaml.safe_load(self.yaml_input)
            return True  # The YAML is valid
        except yaml.YAMLError as e:
            print(f"Invalid YAML string: {e}")
            return False  # The YAML is invalid
            
    def build(self, services: List[Service], tests_path: str, project_name: str):     
        data = {
            "version": "3",
            "services": {},
        }
        for index, service in enumerate(services):
            data["services"][service.name] = {}
            if service.type == "DOCKER":
                data["services"][service.name] = DockerHandler.build(service)
            elif service.type == "DB" and service.db_type == "POSTGRESS":
                data["services"][service.name] = PostgresHandler.build(service, index)
            elif service.type == "FUNCTION" and service.lang_type == "NODEJS":
                data["services"][service.name] = NodeFunctionHandler.build(service, index)
            elif service.type == "GCP" and service.service_type == "FIRESTORE":
                data["services"][service.name] = FirestoreHandler.build(index)
            elif service.type == "GCP" and service.service_type == "STORAGE":
                data["services"][service.name] = StorageHandler.build(service, index)
                data["services"]["gcs_storage_proxy"] = StorageProxy.build(service, index)
            elif service.type == "GCP" and service.service_type == "PUBSUB":
                data["services"][service.name] = PubSubHandler.build(project_name, service, index)             
                data["services"]["pub_sub_initiator"] = PubSubHandler.build_initiator(project_name, service, index)
                #pass
            else:    
                raise ValueError("Service type not supported")
        if tests_path:
            data["services"]["tests"] = PythonTestHandler.build(project_name, services, tests_path)    
        self.yaml_input = yaml.dump(data)
        print(f"{self.yaml_input}") 
               
                            
    def run(self, mode: str):
        
        if not self.validate_yaml_input():
            print("YAML validation failed. Aborting operation.")
            return
        
        try:
            if mode == 'detached':
                # Start a subprocess that runs docker-compose with stdin as the input
                process = subprocess.Popen(['docker', 'compose', '-f', '-', 'up', '--build', '-d'], stdin=subprocess.PIPE, text=True)
            else:
                # Start a subprocess that runs docker-compose with stdin as the input
                process = subprocess.Popen(['docker', 'compose', '-f', '-', 'up', '--build'], stdin=subprocess.PIPE, text=True)
            # Write the YAML string to docker-compose's stdin
            process.communicate(input=self.yaml_input)
            if process.returncode == 0:
                print("Docker Compose up executed successfully.")
            else:
                print(f"Failed to run Docker Compose up with return code {process.returncode}")
            
            logs_process = subprocess.Popen(['docker', 'compose', '-f', '-', 'logs', '-f', 'tests'], stdin=subprocess.PIPE,stdout=subprocess.PIPE, text=True)
            
            #time.sleep(5)  # Give it a moment to terminate
            # Write the YAML string to docker-compose's stdin
            stdout, _ = logs_process.communicate(input=self.yaml_input)
            if logs_process.returncode == 0:
                print("Docker Compose logs executed successfully.")
            else:
                print(f"Failed to run Docker Compose logs with return code {logs_process.returncode}")
            exit_code = None
            # Parse the output for service exit codes
           
            if stdout:
                for line in stdout.splitlines():
                    if "exited with code" in line:
                        exit_code = line.split("exited with code")[1].strip()
                        break
            else:
                print("Stdout is empty.")
            print(f"{stdout}")
    
            if exit_code in ["0"]:
                self.test_results = True
            else:
                self.test_results = False

            # Stop the subprocess
            logs_process.terminate()  # Try to terminate the process gently

            # Optionally, wait a bit to see if the process terminates gracefully
            #time.sleep(2)  # Give it a moment to terminate

            if logs_process.poll() is None:  # Check if the process has indeed terminated
                print("Process did not terminate, killing it.")
                logs_process.kill()  # Forcefully kill the process if it didn't terminate
            else:
                print("Process terminated successfully.")                
        except Exception as e:
            print(f"An error occurred: {e}")
        

    def shutdown_services(self):
        try:
            # Execute docker compose down command
            result = subprocess.run(['docker', 'compose', '-f', '-', 'down'], input=self.yaml_input, text=True, capture_output=True)
            
            if result.returncode == 0:
                print("Successfully shut down all services.")
            else:
                print(f"Failed to shut down services with return code {result.returncode}. Error: {result.stderr}")
        except Exception as e:
            print(f"An error occurred while trying to shut down services: {e}")
                
    def get_test_results(self) -> bool:
        return self.test_results        



from typing import Dict, List, Optional

from pydantic import BaseModel
from enum import Enum

class GCPServiceType(str, Enum):
    FIRESTORE = 'FIRESTORE'
    PUBSUB = 'PUBSUB'
    STORAGE = 'STORAGE'

class ServiceType(str, Enum):
    DOCKER = 'DOCKER'
    GCP = 'GCP'
    FUNCTION = 'FUNCTION'
    DB = 'DB'    

class DBType(str, Enum):
    POSTGRESS = 'POSTGRESS'
                
class LANGType(str, Enum):
    NODEJS = 'NODEJS'
    PYTHON = 'PYTHON'

class GCPSubscription(BaseModel):
    name: str
    push_endpoint: str

class GCPTopic(BaseModel):
    name: str
    subscriptions: List[GCPSubscription]

class GCPBucket(BaseModel):
    name: str
    location: Optional[str] = None
    
class Service(BaseModel):
    id: Optional[str] = None
    name: str
    type: ServiceType
    downstream: bool
    platform: Optional[str] = "linux/amd64" # Default value for non linux platforms
    env_vars: Optional[Dict[str, str | int]] = None
    volumes: Optional[List[Dict[str, str]]] = None
    container_command: Optional[str] = None
    container_command_args: Optional[str] = None
    dependencies: Optional[List[str]] = None

class DockerService(Service):
    image_uri: str
    container_port: int
    host_port: int
    depends_on_services: Optional[List[Service]] = None
        
class GCPService(Service):
    service_type: GCPServiceType
    container_port: Optional[int] = None
    host_port: Optional[int] = None
    depends_on_services: Optional[List[Service]] = None

class GCPStorageService(GCPService):
    buckets: List[GCPBucket]
    
class GCPPubSubService(GCPService):
    topics: List[GCPTopic]
    
class FunctionService(Service):
    code_repo: str
    code_branch: str
    code_folder: Optional[str] = None
    lang_type: LANGType
    container_port: Optional[int] = None
    host_port: Optional[int] = None
    depends_on_services: Optional[List[Service]] = None

class DBService(Service):
    db_type: DBType
    container_port: Optional[int] = None
    host_port: Optional[int] = None
    sql_init_file: Optional[str] = None
    depends_on_services: Optional[List[Service]] = None
    


from time import time
import yaml
import json
from typing import List, Optional
import subprocess
from pydantic import BaseModel
from platforms.IPlatform import IPlatform
from services.Service import Service
import os

class DockerHandler(): #pragma: no cover

    @classmethod        
    def build(cls, service: Service):     
        service_dict = {}
        if service.platform:
            service_dict["platform"] = service.platform    
        if service.env_vars:
            service_dict["environment"] = service.env_vars    
        if service.container_command:
            service_dict["command"] = service.container_command
        if service.container_command_args:
            service_dict["command_args"] = service.container_command_args        
        if service.volumes:
            volumes = []
            for volume in service.volumes:
                volumes.append(f"{volume['host']}:{volume['container']}")
            service_dict["volumes"] = volumes
        if service.host_port and service.container_port:
            service_dict["ports"] = [f"{service.host_port}:{service.container_port}"]
            service_dict["healthcheck"] = {
                                                                
                                                                    "test": [
                                                                    "CMD-SHELL",
                                                                    "sh",
                                                                    "-c",
                                                                    f"nc -z 127.0.0.1 {service.container_port} || exit 1"
                                                                    ],
                                                                    "interval": "5s",
                                                                    "timeout": "2s",
                                                                    "retries": 3,
                                                                    "start_period": "1s"
                                                                }
            
        else:
            raise ValueError("Host port and container port are required for Docker services.")
        if service.image_uri:
            service_dict["image"] = service.image_uri
        if service.depends_on_services:
            depends_on = {}
            env_update = {}
            for dependency in service.depends_on_services:
                depends_on.update({
                                    dependency.name : {
                                        "condition": "service_healthy"
                                    }
                                    })
                if dependency.type == 'GCP' and dependency.service_type == 'FIRESTORE':
                    env_update.update({
                        "FIRESTORE_EMULATOR_HOST": f"{dependency.name}:8080"
                    })
                elif dependency.type == 'GCP' and dependency.service_type == 'PUBSUB':
                    env_update.update({
                        "PUBSUB_EMULATOR_HOST": f"{dependency.name}:8085"
                    })
                elif dependency.type == 'GCP' and dependency.service_type == 'STORAGE':
                    env_update.update({
                        "STORAGE_EMULATOR_HOST": "http://gcs_storage_proxy:5050"
                    })        
            service_dict["depends_on"] = depends_on
            service_dict["environment"].update(env_update)
        
        return service_dict                    
    


from typing import Dict, List, Optional
from platforms.IPlatform import IPlatform
from pydantic import BaseModel
from services.Service import Service


class MicroTestSuite(BaseModel):
    id: Optional[str]
    name: str
    services: List[Service]
    platform: IPlatform
    tests_path: str

    def _find_service_by_name(self, name):
        for service in self.services:
            if service.name == name:
                return service
        return None
    
    def _set_depends_on_services(self):
        for service in self.services:
            if service.dependencies and hasattr(service, 'depends_on_services'):
                service.depends_on_services = [self._find_service_by_name(dep) for dep in service.dependencies]
                                    
    def run(self, mode: str):
        self._set_depends_on_services()
        self.platform.build(self.services, self.tests_path, self.name)
        self.platform.run(mode)
    
    def stop(self):
        self.platform.shutdown_services()
        
    def test_results(self):
        return self.platform.get_test_results()    
        
    class Config:
        arbitrary_types_allowed = True    


import tempfile
import textwrap
from services.Service import Service
import os
import random
import copy
import json

class StorageProxy(): 
    
    pub_sub_init_config = { 
                            "platform": "linux/amd64",
                            "build": {
                                            "context": "{code_folder}",
                                            "dockerfile": "dockerfile.docker"
                                            },
                            "environment": {
                                                "NODE_TLS_REJECT_UNAUTHORIZED":"0"
                                            },
                            "depends_on": "{gcs_emulator_host}",
                            "ports": "{host_port}:5050",
                            }
    
    docker_file_content = """
# Use the official lightweight Node.js 14 image.
# https://hub.docker.com/_/node
FROM node:14-slim

# Create and change to the app directory.
WORKDIR /usr/src/app

# Copy application dependency manifests to the container image.
# A wildcard is used to ensure both package.json AND package-lock.json are copied.
# Copying this separately prevents re-running npm install on every code change.
COPY ./package*.json ./

# Install production dependencies.
RUN npm install --only=production

# Copy local code to the container image.
COPY . .

# Run the web service on container startup.
CMD ["node", "index.js"]
"""

    init_script_content = """const http = require('http');
const httpProxy = require('http-proxy');

// Create a proxy server with custom application logic
const proxy = httpProxy.createProxyServer({
  secure: false // Disable SSL verification
});

const server = http.createServer(function(req, res) {
  // Check if the request method is GET and the URL does not already include /storage/v1
  if (req.method === 'GET' && !req.url.startsWith('/storage/v1')) {
    req.url = '/storage/v1' + req.url; // Prepend /storage/v1 to the URL for GET requests
  }

  // Proxy the request to the GCS emulator
  proxy.web(req, res, { target: 'https://##gcs_emulator##:4443' });
});

console.log("Listening on port 5050")
server.listen(5050);     
"""

    package_json ={
    "name": "http_proxy",
    "version": "1.0.0",
    "description": "A simple Node.js app running in Docker",
    "main": "index.js",
    "scripts": {
      "start": "node index.js"
    },
    "dependencies": {
      "http-proxy": "^1.18.1"
    }
  }
                                   
    @classmethod
    def _load_config(cls, host_port_index, service_name, service_folder):
        host_port = str(5050 + host_port_index)
        # get filename from file path
        config = copy.deepcopy(cls.pub_sub_init_config)
        config['ports'] = [config['ports'].replace("{host_port}", host_port)]
        config['depends_on'] = [config['depends_on'].replace("{gcs_emulator_host}", service_name)]
        config['build']['context'] = service_folder
        
        return config
    
    @classmethod
    def _create_init_file(cls, service_name, service_folder, init_file_content):
       init_file_content = textwrap.dedent(init_file_content)  # Remove leading spaces
       init_file_content = init_file_content.replace("##gcs_emulator##", service_name) 
       os.makedirs(service_folder, exist_ok=True)
       with open(f"{service_folder}/index.js", "w") as file:
            file.write(init_file_content)    
            file_path = file.name
       os.chmod(file_path, 0o755)  # Set executable permissions
       return file_path    
   
    @classmethod
    def _create_docker_file(cls, service_folder, docker_file_content):
       os.makedirs(service_folder, exist_ok=True)
       os.makedirs(service_folder, exist_ok=True)
       docker_file_content = textwrap.dedent(docker_file_content)  # Remove leading spaces
       with open(f"{service_folder}/dockerfile.docker", "w") as file:
            file.write(docker_file_content)    
            file_path = file.name
       os.chmod(file_path, 0o755)  # Set executable permissions
       return file_path 

    
    @classmethod
    def create_package_json(cls, service_folder):
        with open(f"{service_folder}/package.json", "w") as file:
            file.write(json.dumps(cls.package_json))    
            file_path = file.name
        return file_path
       
    
    @classmethod        
    def build(cls, service:Service, index: int):     
        random_number = random.randint(1000, 9999)
        temp_dir = tempfile.gettempdir()
        service_folder = f"/{temp_dir}/{service.name}_proxy_{random_number}"                         
        cls._create_init_file(service.name, service_folder, cls.init_script_content)            
        cls._create_docker_file(service_folder, cls.docker_file_content)
        cls.create_package_json(service_folder)
        return cls._load_config(index, service.name, service_folder)             




import tempfile
from typing import List
from services.Service import Service
import os
import random
import textwrap
import copy
import shutil

class TestHandler(): 

    default_conf = {
                        "platform": "linux/amd64",
                            "build": {
                                            "context": "{code_folder}",
                                            "dockerfile": "dockerfile.docker"
                                            },
                            "depends_on": {}
                        }
    
    
    
    @classmethod
    def _create_docker_file(cls, code_folder, docker_file_content):
       os.makedirs(code_folder, exist_ok=True)
       docker_file_content = docker_file_content.replace("{code_folder}", code_folder) 
       docker_file_content = textwrap.dedent(docker_file_content)  # Remove leading spaces
       with open(f"{code_folder}/dockerfile.docker", "w") as file:
            file.write(docker_file_content)    
            file_path = file.name
       os.chmod(file_path, 0o755)  # Set executable permissions
       return file_path    

            
                                
    @classmethod
    def _load_default_config(cls, code_folder, temp_sh_file):
        # get filename from file path
        config = copy.deepcopy(cls.default_conf)
        config['build']['context'] = code_folder
            
        return config
    
    
    @classmethod        
    def _common_build(cls, dependencies: List, code_folder: str, docker_file_path: str):     
        service_dict = {}
        depends_on = {} 
        service_dict = cls._load_default_config(code_folder, docker_file_path)    
        for dependency in dependencies:
                depends_on.update({
                                    dependency : {
                                        "condition": "service_healthy"
                                    }
                                    })
        service_dict["depends_on"] = depends_on
        return service_dict         

    @classmethod
    def _get_dependencies(cls, services: List[Service]):
        dependencies = []
        for service in services:
            if service.downstream:
                dependencies.append(service.name)
        #dependencies.append("pub_sub_initiator")     
        return dependencies
    
class PythonTestHandler(TestHandler):                
    
    docker_file_content = """
FROM --platform=linux/amd64 python:3.8
WORKDIR /app
COPY requirements.txt .
RUN pip install -r requirements.txt
COPY ./* /app
CMD ["python3", "main.py"]
"""

   
    @classmethod        
    def build(cls, project_name, services: List[Service], test_path: str):     
        service_dict = {}
        
        random_number = random.randint(1000, 9999)
        temp_dir = tempfile.gettempdir()
        test_folder = f"/{temp_dir}/{project_name}_tests_{random_number}"
        # copy the contest of the tests_path to the test_folder
        shutil.copytree(test_path, test_folder, dirs_exist_ok=True)
                
        docker_file_path = cls._create_docker_file(test_folder, cls.docker_file_content)  
        dependencies = cls._get_dependencies(services)  
        service_dict = cls._common_build(dependencies, test_folder, docker_file_path)
        
        return service_dict  

    
    


from typing import Dict, List, Optional
from platforms.IPlatform import IPlatform
from pydantic import BaseModel
from platforms.docker_compose.DockerComposePlatform import DockerComposePlatform
from services.Service import DBService, DockerService, FunctionService, GCPPubSubService, GCPService, GCPStorageService, Service, ServiceType
from test_suites.MicroTestSuite import MicroTestSuite
from enum import Enum

class PlatformType(str, Enum):
    DOCKER = 'DOCKER-COMPOSE'
    GCP = 'GCP'
    
    
class MicroTestProject(BaseModel):
    name: str
    platform: PlatformType
    micro_test_suite: Optional[MicroTestSuite]=None
    tests_path: str
    services: List[Dict]
    
    def _build_services(self):
        services = []
        
        for service in self.services:
            if service['type'] == ServiceType.DOCKER:
                service = DockerService(**service)
            elif service['type'] == 'GCP' and service['service_type'] == 'STORAGE':
                    service = GCPStorageService(**service)
            elif service['type'] == 'GCP' and service['service_type'] == 'FIRESTORE':    
                    service = GCPService(**service)
            elif service['type'] == 'GCP' and service['service_type'] == 'PUBSUB':    
                    service = GCPPubSubService(**service)
            elif service['type'] == 'FUNCTION':
                service = FunctionService(**service)    
            elif service['type'] == 'DB':
                service = DBService(**service)    
            else:
                raise ValueError("Service type not supported")
            services.append(service)
        return services        
    
    def _build_platform(self):
        if self.platform == PlatformType.DOCKER:
            return DockerComposePlatform()
        elif self.platform == PlatformType.GCP:
            pass
        else:
            raise ValueError("Platform not supported")
        
    def run_test(self, mode: str):        
        test_suite = MicroTestSuite(id="1", name=self.name, 
                                    services=self._build_services(), 
                                    platform=self._build_platform(), 
                                    tests_path=self.tests_path)
        test_suite.run(mode)
        results = test_suite.test_results()
        test_suite.stop()
        return results


import os
from typing import Dict, List
import click
import yaml
import json

from test_suites.MicroTestProject import MicroTestProject

MICROHUB_FILE_EXTENSION = '.microhub'
ACTIVE_PROJECT_FILE = '.mhap'
valid_service_types = ['GCP', 'DOCKER', 'FUNCTION', 'DB']
valid_gcp_services = ['PUBSUB', 'FIRESTORE', 'STORAGE']
valid_db_names = ['POSTGRESS']
valid_platform_types = ['DOCKER-COMPOSE', 'GCP']
valid_code_languages = ['NODEJS', 'PYTHON']

def is_valid_env_file(file_path: str) -> bool:
    """
    Checks if the given file is a valid environment file.

    Args:
        file_path (str): The path to the environment file.

    Returns:
        bool: True if the file is valid, False otherwise.
    """
    if not os.path.isfile(file_path):
        click.echo('The specified file does not exist.')
        return False

    try:
        with open(file_path, 'r') as file:
            for line in file:
                if line.strip() and not line.startswith('#'):
                    if '=' not in line:
                        click.echo('The specified file is not a valid environment file.')
                        return False
        return True
    except Exception as e:
        click.echo(f'An error occurred while reading the file: {e}')
        return False

def load_env_file_to_dict(file_path: str) -> dict:
    """
    Loads the content of the environment file into a dictionary.

    Args:
        file_path (str): The path to the environment file.

    Returns:
        dict: A dictionary containing the key-value pairs from the environment file.
    """
    env_dict = {}
    try:
        with open(file_path, 'r') as file:
            for line in file:
                if line.strip() and not line.startswith('#'):
                    key, value = line.strip().split('=', 1)
                    env_dict[key] = value
    except Exception as e:
        click.echo(f'An error occurred while reading the file: {e}')
    return env_dict

def prompt_for_env_vars() -> dict:
    """
    Prompts the user to input environment variables manually.

    Returns:
        dict: A dictionary containing the key-value pairs entered by the user.
    """
    env_vars = {}
    while True:
        key = click.prompt('Please enter the environment variable name (or leave blank to finish)', default='', show_default=False)
        if key == '':
            break
        value = click.prompt(f'Please enter the value for {key}')
        env_vars[key] = value
    return env_vars

def create_env_vars() -> dict:
    """
    Creates environment variables either from a file or by getting the variable name and value from the input.

    Returns:
        dict: A dictionary containing the environment variables.
    """
    while True:
        env_file = click.prompt('Please enter the path of env file (or leave blank to enter variables manually)', default='', show_default=True)
        if env_file == '':
            return prompt_for_env_vars()
        if is_valid_env_file(env_file):
            return load_env_file_to_dict(env_file)

def prompt_for_selection(prompt_message: str, options: List, no_selection: bool=True) -> str|None:
    """
    Prompts the user to select an option from a list of options.

    Args:
        prompt_message (str): The message to display to the user.
        options (list): The list of options to choose from.
        no_selection (bool): Whether to include an option for no selection.

    Returns:
        The selected option, or None if the selection is invalid.
    """
    
    while True:
        # Display the list of options with indices
        click.echo(prompt_message)
        for i, option in enumerate(options):
            click.echo(f'{i}: {option}')
        if no_selection:    
            click.echo(f'{len(options)}: No Selection')
        # Prompt the user to enter the index of the desired option
        option_index = click.prompt('Enter the number corresponding to your choice', type=int)

        # Validate the user's input and return the corresponding option
        if 0 <= option_index < len(options):
            return options[option_index]
        elif no_selection:
            return None
        else:
            click.echo('Invalid selection. Select a valid option.')


def folder_exists(folder: str) -> bool:
    # check if the folder path exists
    if not os.path.isdir(folder):
        return False
    return True
    
    
def set_project_name(name: str) -> str:
    # doc string comments
    """
    Replaces spaces in the project name with underscores.
    
    Args:
        name (str): The project name.
        
    Returns:
        str: The project name with spaces replaced by underscores.
    """
    # replace the spaces with underscores
    name = name.replace(' ', '_')
    return name

def read_active_project() -> str:
    active_project = None
    if os.path.isfile(ACTIVE_PROJECT_FILE):
        with open(ACTIVE_PROJECT_FILE, 'r') as file:
            active_project = file.read().strip()
    return active_project

def add_services() -> List[Dict]|None:
    """
    Prompts the user to add services to the project configuration.
    
    Returns:
        List[Dict]: A list of dictionaries containing the service configurations.
    """
    add_service = True
    services = []
    while add_service:
        service = {}
        service['name'] = click.prompt('Please enter the service name')

        # Use the generic function to prompt for service type
        service['type'] = prompt_for_selection('Please select the project type:', valid_service_types)
        service['downstream'] = click.confirm('Is it a service that your test could interact with?')
       
        if service['type'] == 'DOCKER':
            service['image_uri'] = click.prompt('Please enter the path of the image')
            service['container_port'] = click.prompt('Please enter the container port', type=int) 
            service['host_port'] = click.prompt('Please enter the host port', type=int)
            service['volumes'] = add_service_volumes()
            service['container_command'] = click.prompt('Please enter the container_command', default='', show_default=True)
            service['container_command_args'] = click.prompt('Please enter the container_command_args', default='', show_default=True)
            service['env_vars'] = create_env_vars()
        elif service['type'] == 'GCP':
            # Use the generic function to prompt for service type
            service['service_type'] = prompt_for_selection('Please select the gcp service name:', valid_gcp_services)
            if service['service_type'] == 'GCS':
                service['buckets'] = []
                add_bucket = True
                while add_bucket:
                    bucket = {}
                    bucket['name'] = click.prompt('Please enter the bucket name')
                    bucket['location'] = click.prompt('Please enter the bucket location', default='', show_default=True)
                    service['buckets'].append(bucket)
                    add_bucket = click.confirm('Do you want to add another bucket?')
            elif service['service_type'] == 'PUBSUB':
                service['topics'] = []
                add_topic = True
                while add_topic:
                    topic = {}
                    topic['name'] = click.prompt('Please enter the topic name')
                    topic['subscriptions'] = []
                    add_subscription = True
                    while add_subscription:
                        subscription = {}
                        subscription['name'] = click.prompt('Please enter the subscription name')
                        subscription['push_endpoint'] = click.prompt('Please enter the push end point')
                        topic['subscriptions'].append(subscription)
                        add_subscription = click.confirm('Do you want to add another subscription?')
                    service['topics'].append(topic)
                    add_topic = click.confirm('Do you want to add another topic?')        
        elif service['type'] == 'FUNCTION':
            service['code_repo'] = click.prompt('Please enter the repo of the function')    
            service['code_branch'] = click.prompt('Please enter the branch name of the repo')    
            service['code_folder'] = click.prompt('Please enter the folder name of the code', default='', show_default=True)
            service['lang_type'] = prompt_for_selection('Please select the function code language:', valid_code_languages,
                                                        no_selection=False)
            service['env_vars'] = create_env_vars()
        elif service['type'] == 'DB':
            # Use the generic function to prompt for service type
            service['db_type'] = prompt_for_selection('Please select the DB name:', valid_db_names)
            service['sql_init_file'] = click.prompt('Please enter the path of the sql init file') 
               
        services.append(service)
        add_service = click.confirm('Do you want to add another service?')
    return services    

def add_service_volumes() -> List[Dict]:
    """
    Prompts the user to add volumes to the service configuration.
    
    Args:
        service (Dict): A dictionary containing the service configuration.
        
    Returns:
        Dict: A dictionary containing the service configuration with volumes added.
    """
    add_volume = True
    volumes = []
    while add_volume:
        volume = {}
        volume['container'] = click.prompt('Please enter the container volume')
        volume['host'] = click.prompt('Please enter the host volume')
        volumes.append(volume)
        add_volume = click.confirm('Do you want to add another volume?')
    return volumes

def add_services_dependencies(services: List[Dict]) -> List[Dict]:
    """
    Prompts the user to add dependencies between services.
    
    Args:
        services (List[Dict]): A list of dictionaries containing the service configurations.
        
    Returns:
        List[Dict]: A list of dictionaries containing the service configurations with dependencies added.
    """
    all_services = [service['name'] for service in services]
    for service in services:
        # only add dependencies for DOCKER and FUNCTION services
        if service['type'] not in  ['DOCKER', 'FUNCTION']:
            continue
        dependencies = []
        all_but_current = [s for s in all_services if s != service['name']]
        add_service_dependency = True
        while add_service_dependency:
            dependency = prompt_for_selection(f'Adding service dependencies for {service['name']}', all_but_current)
            if dependency:
                dependencies.append(dependency)
                add_service_dependency = click.confirm('Do you want to add another dependency?')
            else:
                add_service_dependency = False    
        if 'dependencies' not in service:
            service['dependencies'] = dependencies
        else:
            service['dependencies'].extend(dependencies)
    return services

@click.command()
def create_project():
    """
    Creates a new project configuration file.
    """
    
    project = {}
    
    # Ask for project name
    project['name'] = set_project_name(click.prompt('Please enter the project name'))
    project['tests_path'] = click.prompt('Please enter the path of the test folder')

    while not folder_exists(project['tests_path']):
        project['tests_path'] = click.prompt('The specified folder does not exist. Please enter a valid path')
        
    project['platform'] = prompt_for_selection('Please select the project platform:', 
                                                   valid_platform_types, 
                                                   no_selection=False)
    project['services'] = add_services()
    project['services'] = add_services_dependencies(project['services'])
    
    # Convert the project dictionary to a YAML string
    project_yaml = yaml.dump(project, default_flow_style=False)

    # Save the YAML string to a file
    with open(f'{project['name']}{MICROHUB_FILE_EXTENSION}', 'w') as file:
        file.write(project_yaml)

    click.echo(f'Project {project['name']} created successfully!')

@click.command()
@click.option('--name', required=True, help='project name')
def update_project(name):
    """
    Update the given project.
    
    Args:
    
    name (str): The name of the project to set as active.
    """
    # search for a yaml file with the project name
    project_file = f'{name}{MICROHUB_FILE_EXTENSION}'
    if not os.path.isfile(project_file):
        click.echo(f'{name} PROJECT DOES NOT EXIST')
        return
    with open(project_file, 'r+') as file:
        project = yaml.safe_load(file)
        click.echo(json.dumps(project, indent=4))
        updated_services = add_services()
        project['services'].extend(updated_services)
        project['services'] = add_services_dependencies(project['services'])
        project_yaml = yaml.dump(project, default_flow_style=False)
        file.seek(0)
        file.write(project_yaml)
        file.truncate()
        click.echo(f'Project {project['name']} updated successfully!')
        
@click.command()
@click.option('--name', required=True, help='project name')
def set_active_project(name):
    """
    Sets the active project.
    
    Args:
    
    name (str): The name of the project to set as active.
    """
    
    # search for a yaml file with the project name
    if not os.path.isfile(f'{name}{MICROHUB_FILE_EXTENSION}'):
        click.echo(f'{name} PROJECT DOES NOT EXIST')
        return
	# Write the active project name to the .mhap file
    with open(ACTIVE_PROJECT_FILE, 'w') as file:
        file.write(name)
    click.echo(f'Active project set to {name}')
    
@click.command()
def list_projects():
    
    """
    Lists all the projects.
    """
    
    project_files = [f for f in os.listdir('.') if f.endswith(MICROHUB_FILE_EXTENSION)]
    active_project = read_active_project()
    for project_file in project_files:
        with open(project_file, 'r') as file:
            project = yaml.safe_load(file)
            if active_project == project['name']:
                click.echo(f'{project['name']} (Active)')
            else:
                click.echo(project['name'])    
            
@click.command()
@click.option('--name', required=True, help='project name')
def delete_project(name):
    # delete the yaml file with name the project name
    if not os.path.isfile(f'{name}{MICROHUB_FILE_EXTENSION}'):
        click.echo(f'{name} PROJECT DOES NOT EXIST')
        return
    os.remove(f'{name}{MICROHUB_FILE_EXTENSION}')
    # if the project is the active project, remove the active project file
    active_project = read_active_project()
    if active_project == name:
        os.remove(ACTIVE_PROJECT_FILE)
    click.echo(f'{name} PROJECT DELETED SUCCESSFULLY')
    

@click.command()
@click.option('--project-name', required=False, help='project name')
@click.option('--mode', type=click.Choice(['attached', 'detached']), default='detached', required=False, help='Mode of operation')
def run_test(project_name=None, mode='detached'):
    
    """
    Runs the tests for the specified project.
    
    Args:
    project_name (str): The name of the project to run the tests for.
    """
    
    if not project_name:
        project_name = read_active_project()
        if not project_name:
            click.echo('No active project set. Please specify a project name.')
            return
    project_file = f'{project_name}{MICROHUB_FILE_EXTENSION}'
    if not os.path.isfile(project_file):
        click.echo(f'{project_name} project does not exist.')
        return
    with open(project_file, 'r') as file:
        project = yaml.safe_load(file)
        click.echo(json.dumps(project, indent=4))
        # Run the tests
        micro_test_project = MicroTestProject(**project)
        results = micro_test_project.run_test(mode)
        click.echo(f'results for project {project_name} is {results}')
        
        


from time import time
import yaml
from typing import List, Optional
import subprocess
from pydantic import BaseModel
from platforms.IPlatform import IPlatform
from services.Service import Service
import os

class DCPlatform(BaseModel, IPlatform): #pragma: no cover

    yaml_input: Optional[str] = None
    test_results: Optional[bool] = None
    
    def validate_yaml_input(self) -> bool:
        try:
            yaml.safe_load(self.yaml_input)
            return True  # The YAML is valid
        except yaml.YAMLError as e:
            print(f"Invalid YAML string: {e}")
            return False  # The YAML is invalid
        
    def build(self, services: List[Service], tests_path: str):
        print(f"Current working directory: {os.getcwd()}")
        data = {
            "version": "3",
            "services": {},
            # "networks": {
            #     "simple_feature_management_network": {
            #         "external": True
            #     }
            # }
        }
        depends_on_list = {}
        for service in services:
            if service.downstream:
                depends_on_list.update({
                                        service.name: 
                                            {
                                                "condition": "service_healthy"
                                            }
                                        })
            data["services"][service.name] = {
                "volumes": service.volumes
            }
            if service.platform:
                data["services"][service.name]["platform"] = service.platform
            if service.build:
                data["services"][service.name]["build"] = service.build
            if service.ports:
                data["services"][service.name]["ports"] = [service.ports]
            if service.environment:
                data["services"][service.name]["environment"] = service.environment
            if service.image_uri:
                data["services"][service.name]["image"] = service.image_uri
            if service.depends_on:
                data["services"][service.name]["depends_on"] = service.depends_on
            if service.health_check:
                data["services"][service.name]["healthcheck"] = service.health_check    
            if service.command:
                data["services"][service.name]["command"] = service.command    
        if tests_path:
            data["services"]["tests"] = {
                "build": {
                    "context": f"{tests_path}",
                    "dockerfile": "dockerfile.docker"
                    },
                "depends_on": depends_on_list
            }    
        self.yaml_input = yaml.dump(data)
        print(f"{self.yaml_input}")
        
    def run(self):
        
        if not self.validate_yaml_input():
            print("YAML validation failed. Aborting operation.")
            return
        
        try:
            # Start a subprocess that runs docker-compose with stdin as the input
            process = subprocess.Popen(['docker', 'compose', '-f', '-', 'up', '--build'], stdin=subprocess.PIPE, text=True)
            # Write the YAML string to docker-compose's stdin
            process.communicate(input=self.yaml_input)
            if process.returncode == 0:
                print("Docker Compose up executed successfully.")
            else:
                print(f"Failed to run Docker Compose up with return code {process.returncode}")
            
            logs_process = subprocess.Popen(['docker', 'compose', 'logs', '-f', 'tests'], stdout=subprocess.PIPE, text=True)
            
            #time.sleep(5)  # Give it a moment to terminate
            # Write the YAML string to docker-compose's stdin
            stdout, _ = logs_process.communicate()
            if logs_process.returncode == 0:
                print("Docker Compose up executed successfully.")
            else:
                print(f"Failed to run Docker Compose up with return code {logs_process.returncode}")
            exit_code = None
            # Parse the output for service exit codes
           
            if stdout:
                for line in stdout.splitlines():
                    if "exited with code" in line:
                        exit_code = line.split("exited with code")[1].strip()
                        break
            else:
                print("Stdout is empty.")
            print(f"{stdout}")
    
            if exit_code in ["0"]:
                self.test_results = True
            else:
                self.test_results = False

            # Stop the subprocess
            logs_process.terminate()  # Try to terminate the process gently

            # Optionally, wait a bit to see if the process terminates gracefully
            #time.sleep(2)  # Give it a moment to terminate

            if logs_process.poll() is None:  # Check if the process has indeed terminated
                print("Process did not terminate, killing it.")
                logs_process.kill()  # Forcefully kill the process if it didn't terminate
            else:
                print("Process terminated successfully.")                
        except Exception as e:
            print(f"An error occurred: {e}")
    
    def shutdown_services(self):
        try:
            # Execute docker compose down command
            result = subprocess.run(['docker', 'compose', '-f', '-', 'down'], input=self.yaml_input, text=True, capture_output=True)
            
            if result.returncode == 0:
                print("Successfully shut down all services.")
            else:
                print(f"Failed to shut down services with return code {result.returncode}. Error: {result.stderr}")
        except Exception as e:
            print(f"An error occurred while trying to shut down services: {e}")
            
    def get_test_results(self) -> bool:
        return self.test_results        
